/*
 *	dirtime.c
 */

#if defined(UNICODE) && !defined(_UNICODE)
#define _UNICODE
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <string.h>
#include <tchar.h>
#include "vwinfunc.h"
#include "dirtime.h"


BOOL WINAPI SetDirectoryTime(LPCTSTR path, const FILETIME *pCreationTime,
		const FILETIME *pLastAccessTime, const FILETIME *pLastWriteTime)
{
	HANDLE hFile;
	DWORD attr, ret;
	DWORD flag = FILE_ATTRIBUTE_NORMAL;
	
#if defined(UNICODE) && !defined(NOWIN9X)
	if (IsWindows9x()) {
		char buf[MAX_PATH];
		
		if (!WideCharToMultiByte(CP_ACP, WC_SEPCHARS, path, -1,
				buf, sizeof(buf), NULL, NULL)) {
			return FALSE;
		}
		return SetDirectoryTimeA(buf, pCreationTime, pLastAccessTime,
					pLastWriteTime);
	}
#endif
	
	attr = GetFileAttributes(path);
	if (attr == -1) {
		return FALSE;
	}
	if (attr & FILE_ATTRIBUTE_DIRECTORY) {
#ifndef UNICODE
		if (IsWindows9x()) {
			return SetDirectoryTime9x(path, pCreationTime, pLastAccessTime,
					pLastWriteTime);
		} else
#endif
		{
			flag |= FILE_FLAG_BACKUP_SEMANTICS;
		}
	}
	
	hFile = CreateFile(path, FILE_WRITE_ATTRIBUTES,
			FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, flag, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	ret = SetFileTime(hFile, pCreationTime, pLastAccessTime,
			pLastWriteTime);
	CloseHandle(hFile);
	
	return ret;
}

#ifndef UNICODE
BOOL SetDirectoryTime9x(const char *path, const FILETIME *pCreationTime,
		const FILETIME *pLastAccessTime, const FILETIME *pLastWriteTime)
{
	char fullpath[MAX_PATH];
	WORD dosdate, dostime, millisec;
	BOOL ret = TRUE;
	
	if (!GetFullPathName(path, sizeof(fullpath), fullpath, NULL)) {
		return FALSE;
	}
	
	if (pCreationTime != NULL) {
		ktFileTimeToDosDateTime(pCreationTime,
				&dosdate, &dostime, &millisec, FT_ROUND_NONE);
		ret = SetDirectoryTimeDos(fullpath, SET_CREATION_TIME,
				dosdate, dostime, millisec);
	}
	if (ret && pLastAccessTime != NULL) {
		ktFileTimeToDosDateTime(pLastAccessTime,
				&dosdate, &dostime, &millisec, FT_ROUND_UP);
		ret = SetDirectoryTimeDos(fullpath, SET_LAST_ACCESS_TIME,
				dosdate, dostime, millisec);
	}
	if (ret && pLastWriteTime != NULL) {
		ktFileTimeToDosDateTime(pLastWriteTime,
				&dosdate, &dostime, &millisec, FT_ROUND_UP);
		ret = SetDirectoryTimeDos(fullpath, SET_LAST_WRITE_TIME,
				dosdate, dostime, millisec);
	}
	return ret;
}

BOOL SetDirectoryTimeDos(const char *fullpath, int func,
		WORD dosdate, WORD dostime, WORD millisec)
{
	DPMIREGS regs = {0};
	
	regs.eax = 0x7143;
	regs.ebx = func;
	regs.edi = dosdate;
	regs.ecx = dostime;
	regs.esi = millisec;
	regs.edx = (DWORD) fullpath;	// ds:dx = filename
	VWin32PMInt21(&regs);
	return !(regs.flags & 1);
}
#endif /* UNICODE */


BOOL WINAPI GetDirectoryTime(LPCTSTR path, FILETIME *pCreationTime,
		FILETIME *pLastAccessTime, FILETIME *pLastWriteTime)
{
#ifdef USE_INT21
	HANDLE hFile;
	DWORD attr, ret;
	DWORD flag = FILE_ATTRIBUTE_NORMAL;
	
#if defined(UNICODE) && !defined(NOWIN9X)
	if (IsWindows9x()) {
		char buf[MAX_PATH];
		
		if (!WideCharToMultiByte(CP_ACP, WC_SEPCHARS, path, -1,
				buf, sizeof(buf), NULL, NULL)) {
			return FALSE;
		}
		return GetDirectoryTimeA(buf, pCreationTime, pLastAccessTime,
					pLastWriteTime);
	}
#endif
	
	attr = GetFileAttributes(path);
	if (attr == -1) {
		return FALSE;
	}
#ifndef UNICODE
	if (IsWindows9x()) {
		return GetDirectoryTime9x(path, pCreationTime, pLastAccessTime,
				pLastWriteTime);
	}
#endif
	if (attr & FILE_ATTRIBUTE_DIRECTORY) {
		flag |= FILE_FLAG_BACKUP_SEMANTICS;
	}
	
	hFile = CreateFile(path, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, flag, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	ret = GetFileTime(hFile, pCreationTime, pLastAccessTime,
			pLastWriteTime);
	CloseHandle(hFile);
	
	return ret;
#else /* USE_INT21 */
	HANDLE hFind;
	WIN32_FIND_DATA wfd;
	
#if defined(UNICODE) && !defined(NOWIN9X)
	if (IsWindows9x()) {
		char buf[MAX_PATH];
		
		if (!WideCharToMultiByte(CP_ACP, WC_SEPCHARS, path, -1,
				buf, sizeof(buf), NULL, NULL)) {
			return FALSE;
		}
		return GetDirectoryTimeA(buf, pCreationTime, pLastAccessTime,
					pLastWriteTime);
	}
#endif
	
	if (_tcschr(path, _T('*')) || _tcschr(path, _T('?'))) {	/* wild cards */
		return FALSE;
	}
	hFind = FindFirstFile(path, &wfd);
	if (hFind == INVALID_HANDLE_VALUE) {
		return FALSE;
	}
	FindClose(hFind);
	
	if (pCreationTime != NULL) {
		*pCreationTime = wfd.ftCreationTime;
	}
	if (pLastAccessTime != NULL) {
		*pLastAccessTime = wfd.ftLastAccessTime;
	}
	if (pLastWriteTime != NULL) {
		*pLastWriteTime = wfd.ftLastWriteTime;
	}
	return TRUE;
#endif /* USE_INT21 */
}

#ifndef UNICODE
BOOL GetDirectoryTime9x(const char *path, FILETIME *pCreationTime,
		FILETIME *pLastAccessTime, FILETIME *pLastWriteTime)
{
	char fullpath[MAX_PATH];
	WORD dosdate, dostime, millisec;
	BOOL ret = TRUE;
	
	if (!GetFullPathName(path, sizeof(fullpath), fullpath, NULL)) {
		return FALSE;
	}
	
	if (pCreationTime != NULL) {
		ret = GetDirectoryTimeDos(fullpath, GET_CREATION_TIME,
				&dosdate, &dostime, &millisec);
		if (ret) {
			ktDosDateTimeToFileTime(dosdate, dostime, millisec,
					pCreationTime);
		}
	}
	if (ret && pLastAccessTime != NULL) {
		ret = GetDirectoryTimeDos(fullpath, GET_LAST_ACCESS_TIME,
				&dosdate, &dostime, &millisec);
		if (ret) {
			ktDosDateTimeToFileTime(dosdate, dostime, millisec,
					pLastAccessTime);
		}
	}
	if (ret && pLastWriteTime != NULL) {
		ret = GetDirectoryTimeDos(fullpath, GET_LAST_WRITE_TIME,
				&dosdate, &dostime, &millisec);
		if (ret) {
			ktDosDateTimeToFileTime(dosdate, dostime, millisec,
					pLastWriteTime);
		}
	}
	return ret;
}

BOOL GetDirectoryTimeDos(const char *fullpath, int func,
		WORD *dosdate, WORD *dostime, WORD *millisec)
{
	DPMIREGS regs = {0};
	
	regs.eax = 0x7143;
	regs.ebx = func;
	regs.edx = (DWORD) fullpath;	// ds:dx = filename
	VWin32PMInt21(&regs);
	if (regs.flags & 1) {
		return FALSE;
	}
	if (dosdate != NULL)
		*dosdate = (WORD) regs.edi;
	if (dostime != NULL)
		*dostime = (WORD) regs.ecx;
	if (millisec != NULL)
		*millisec = (WORD) regs.esi;
	return TRUE;
}
#endif /* UNICODE */





#ifndef NO_TIMECONV

BOOL WINAPI ktDosDateTimeToFileTime(WORD dosdate, WORD dostime, WORD millisec,
		FILETIME *pFileTime)
{
	SYSTEMTIME LocalTime;
	
	if (!ktDosDateTimeToLocalTime(dosdate, dostime, millisec, &LocalTime))
		return FALSE;
	if (!ktLocalTimeToFileTime(&LocalTime, pFileTime))
		return FALSE;
	return TRUE;
}

BOOL WINAPI ktFileTimeToDosDateTime(const FILETIME *pFileTime,
		WORD *pdosdate, WORD *pdostime, WORD *pmillisec, int roundmode)
{
	SYSTEMTIME LocalTime;
	FILETIME rtime;
	const FILETIME *pftime = pFileTime;
	
	if (roundmode != FT_ROUND_NONE) {
		int n = (roundmode > 0) ? (20000000 - 1) : 0;
		((ULARGE_INTEGER *) &rtime)->QuadPart =
				(((ULARGE_INTEGER *) pFileTime)->QuadPart + n)
					/ 20000000 * 20000000;
		pftime = &rtime;
	}
	
	if (!ktFileTimeToLocalTime(pftime, &LocalTime))
		return FALSE;
	if (!ktLocalTimeToDosDateTime(&LocalTime, pdosdate, pdostime, pmillisec))
		return FALSE;
	return TRUE;
}

BOOL WINAPI ktDosDateTimeToLocalTime(WORD dosdate, WORD dostime, WORD millisec,
		SYSTEMTIME *pLocalTime)
{
	if (pLocalTime) {
		WORD sec = (WORD) (millisec / 1000);
		
		pLocalTime->wYear   = (WORD) ((dosdate >> 9) + 1980);
		pLocalTime->wMonth  = (WORD) ((dosdate >> 5) & 0x0f);
		pLocalTime->wDay    = (WORD) (dosdate & 0x1f);
		pLocalTime->wHour   = (WORD) (dostime >> 11);
		pLocalTime->wMinute = (WORD) ((dostime >> 5) & 0x3f);
		pLocalTime->wSecond = (WORD) ((dostime & 0x1f) * 2 + sec);
		pLocalTime->wMilliseconds = (WORD) (millisec - sec * 1000);
	}
	return TRUE;
}

BOOL WINAPI ktLocalTimeToDosDateTime(const SYSTEMTIME *pLocalTime,
		WORD *pdosdate, WORD *pdostime, WORD *pmillisec)
{
	WORD sec = (WORD) (pLocalTime->wSecond / 2);
	
	if (pdosdate) {
		/* YYYYYYYMMMMDDDDD */
		*pdosdate = (WORD) ((pLocalTime->wYear - 1980) << 9
				| pLocalTime->wMonth << 5
				| pLocalTime->wDay);
	}
	if (pdostime) {
		/* HHHHHMMMMMMSSSSS */
		*pdostime = (WORD) (pLocalTime->wHour << 11
				| pLocalTime->wMinute << 5
				| sec);
	}
	if (pmillisec) {
		*pmillisec = (WORD) ((pLocalTime->wSecond - sec * 2) * 1000
				+ pLocalTime->wMilliseconds);
	}
	return TRUE;
}

BOOL WINAPI ktLocalTimeToFileTime(const SYSTEMTIME *pLocalTime,
		FILETIME *pFileTime)
{
	FILETIME LocalFileTime;
	
	if (!SystemTimeToFileTime(pLocalTime, &LocalFileTime))
		return FALSE;
	if (!LocalFileTimeToFileTime(&LocalFileTime, pFileTime))
		return FALSE;
	return TRUE;
}

BOOL WINAPI ktFileTimeToLocalTime(const FILETIME *pFileTime,
		SYSTEMTIME *pLocalTime)
{
	FILETIME LocalFileTime;
	
	if (!FileTimeToLocalFileTime(pFileTime, &LocalFileTime))
		return FALSE;
	if (!FileTimeToSystemTime(&LocalFileTime, pLocalTime))
		return FALSE;
	return TRUE;
}

#endif /* NO_TIMECONV */
